# TODO: module for creating simple GUI components and applications
import pyautogui
import klsframe.utilities.utils as klsutils


def confirm_yes_no(selection, default=True):
    if not isinstance(default, bool):
        raise ValueError("Unknown value for parameter 'default'")
    elif default:
        hint = f'[Y/n]'
        butt1 = 'Yes'
        butt2 = 'No'
    else:
        hint = f'[N/y]'
        butt1 = 'No'
        butt2 = 'Yes'
    opt = user_confirmation(title='Confirm selection',
                            text=f"You selected:\n\n{selection}\n\nConfirm selection? {hint}\t",
                            button1=butt1, button2=butt2)
    if opt == 'Yes':
        return True
    elif opt == 'No':
        return False
    else:
        print(f"Option '{opt}' not recognized")


def confirm_continue_or_exit(title="Continue or exit", text="Do you want to continue?",
                             button_continue="continue", button_cancel="exit", timeout=None):
    def say_bye():
        print("You decided to continue...death awaits you...")

    def _user_exit():
        print('[INFO] Execution cancelled by the user. Exiting...')
        pyautogui.alert(title='Exiting', text='Execution cancelled by the user.\nExiting...')
        exit(0)

    user_confirmation(title=title, text=text, button1=button_continue, button2=button_cancel,
                      action1=say_bye, action2=_user_exit, timeout=timeout)


def user_confirmation(title="", text="", button1='continue', button2='cancel',
                      action1=None, action2=None, timeout=None, action_timeout=None):
    """
    Prompts a window to arise a question with two possible answers.
    Useful for example for Continue/cancel prompts

    :param title: Window title
    :param text: Window text content
    :param button1: text to be displayed in button 1
    :param button2: text to be displayed in button 2
    :param action1: callback  executed if button1 is pressed
    :param action2: callback  executed if button1 is pressed
    :param timeout: prompt timeout in seconds
    :param action_timeout: callback (function) executed if timeout is reached
    :return: the pressed button's text
    """
    if timeout is not None:
        ret = pyautogui.confirm(title=title, text=text, buttons=(button1, button2), timeout=timeout * 1000)
    else:
        ret = pyautogui.confirm(title=title, text=text, buttons=(button1, button2))
    if klsutils.equals_ignore_case(ret, button1):
        if action1 is not None:
            action1()
    elif klsutils.equals_ignore_case(ret, button2):
        if action2 is not None:
            action2()
    elif timeout is not None and ret == 'Timeout':
        if action_timeout is not None:
            action_timeout()
    else:
        raise ValueError("Unexpected input value")
    return ret


def prompt(title="", text="", default="", timeout=None, validations=None, error_msg=None,
           confirm=False, allow_empty=True):
    # TODO: implement it like safe input
    """
    Prompts a window for the user to input a text value

    :param title: Prompt title
    :param text: prompt description text
    :param default: Default value for the input
    :param timeout: prompt timeout in seconds
    :param validations: Set of RegExp indicating the possible values of the input, like a whitelist
    :param error_msg: Message to be displayed when the input is not valid
    :param confirm: Enable/disable input confirmation
    :param allow_empty: Enable/disable empty inputs
    :return: The user input as a string or None if the input was cancelled
    """
    if validations is None:
        validations = []
    elif type(validations) is not list:
        validations = [validations]
    if error_msg is None:
        error_msg = "Error. Invalid input"
    while True:
        if timeout is None:
            inval = pyautogui.prompt(title=title, text=text, default=default)
        else:
            inval = pyautogui.prompt(title=title, text=text, default=default, timeout=timeout * 1000)
        if not allow_empty and inval == "":
            pyautogui.alert(title='Input error', text=error_msg)
        elif len(validations) > 0:
            if klsutils.validate(inval, validations):  # Validation passed
                if bool(confirm) and inval is not None:  # Confirmation prompt enabled
                    if confirm_yes_no(inval):
                        return inval  # Confirmation: accepted
                    else:
                        continue  # Confirmation: rejected. Continue the questioning loop
                else:  # Confirmation prompt disabled
                    return inval
            else:  # Validation failed
                pyautogui.alert(title='Input error', text=error_msg)
        else:  # No validations to be passed
            if bool(confirm):  # Confirmation prompt enabled
                if confirm_yes_no(inval):
                    return inval  # Confirmation: accepted
                else:
                    continue  # Confirmation: rejected. Continue the questioning loop
            else:  # Confirmation prompt disabled
                return inval


if __name__ == '__main__':
    # confirm_continue_or_exit()
    print(prompt(title="Testing testing", text="poninedo a prueba tu subwoffer", confirm=True))
    # print(confirm_yes_no('Morir'))



import re
from typing import Union
from klsframe.utilities.utils import validate, text_padding, dict_to_table


def confirm_yes_no(selection=None, default=True, allow_empty=True, shortened=True):
    a = 'yes'
    b = 'no'
    if shortened:
        a = 'y'
        b = 'n'
    if isinstance(default, bool) and default:
        hint = f'[{a.upper()}/{b}]'
    elif isinstance(default, bool) and not default:
        hint = f'[{b.upper()}/{a}]'
    else:
        raise ValueError("Unexpected value for parameter 'default'")

    if selection is None:
        sel_str = ''
    else:
        sel_str = f"You selected:\n{text_padding(selection, decorator='*', decorate_lines='first')}\n"
    while True:
        opt = input(f"{sel_str}>> Do you want to continue? {hint}\t").strip().lower()
        if opt == '':
            if allow_empty:
                return default
            else:
                print("Please provide an answer")
        elif re.match('(no|n)', opt):
            return False
        elif re.match('(yes|y)', opt):
            return True
        else:
            print(f"Option '{opt}' not recognized")


def continue_or_exit():
    # TODO: develop continue_or_exit(). A function that allows the program to continue or stops it
    # Based on confirm_yes_no(), although the 'no' option implies exit()
    if confirm_yes_no():
        return None
    else:
        exit(0)


def safe_string_input(validations=None, prompt=None, error_msg=None, confirm=False, allow_empty=True):
    """
    Safe input for string values. Reads input from terminal, performing some security checks on the input

    Arguments:
        :param validations: Set of RegExp indicating the possible values of the input, like a whitelist
        :param prompt: Custom text to be displayed before reading the input
        :param error_msg: Custom text to be displayed when input does not match valid regex
        :param confirm: Enables/Disables the prompt confirmation after entering a value
        :param allow_empty: Allow/Disallow empty as input
    """
    # Validate the parameters
    if validations is None:
        validations = []
    if error_msg is None:
        error_msg = "Error. Invalid input"
    if prompt is None:
        prompt = "Please introduce a value:"
    while True:
        inval = input(f"{prompt}\t")
        if not allow_empty and inval == "":
            print(f"{error_msg} (Input cannot be empty)")
        elif validate(inval, validations):
            if bool(confirm) and not confirm_yes_no(inval):  # Confirmation prompt enabled
                continue  # Confirmation: rejected. Continue the questioning loop
            else:  # Confirmation prompt disabled, or accepted
                return inval
        else:  # Validation failed
            print(f"{error_msg} (Validation failed)")


def safe_number_input(min_val=None, max_val=None, prompt=None,
                      decimal_digits=None, decimal_sep='.', default=None, **kwargs) -> Union[int, float]:
    """
    Safe input for numeric values.

    :param min_val: Min value allowed. None (-infinite) by default
    :param max_val: Max value allowed. None (infinite) by default
    :param prompt: Custom message for the input prompt
    :param decimal_digits: Decimal digit count. 0 (default) for integer numbers.
        If not None, the input is rounded to ``decimal_digits``
    :param decimal_sep: Character used to separate the decimal part from the integer part.
        Allowed chars: '.' | ','. Note: decimal_sep = '.' <==> thousand_sep = ',' (and vice versa)
    :param default: Default value for the input. If not None, if the user inputs an empty value,
        the default value will be returned. If None, the user input cannot be empty
    :return: The validated and sanitized number input by the user
    """
    if decimal_sep == '.':
        thousand_sep = ','
    elif decimal_sep == ',':
        thousand_sep = '.'
    else:
        raise ValueError("Unexpected value for parameter 'decimal_sep'. Allowed: ,|.")
    if decimal_digits is not None and not isinstance(decimal_digits, int):
        raise TypeError("Unexpected type for parameter 'decimal_digits'. Allowed: int")
    bounds = f"[{min_val}, {max_val}]".replace('[None', '(-infinite').replace('None]', 'infinite)')
    # General number regex
    custom_regex = "^-?([0-9]{1,3}" + thousand_sep + ")*[0-9]+"
    if decimal_digits is None:
        _prompt = f'Please introduce a number within {bounds}'
        custom_regex += "(\\" + decimal_sep + "[0-9]+)?"
    elif decimal_digits > 0:
        custom_regex += "(\\" + decimal_sep + "[0-9]+)?"
        _prompt = f'Please introduce an decimal number (up to {decimal_digits} decimal digits) within {bounds}'
    else:
        _prompt = f'Please introduce an integer number within {bounds}'
    if prompt is not None:
        _prompt = prompt
    # Check parameter `default`
    if 'allow_empty' in kwargs and kwargs['allow_empty'] and isinstance(kwargs['allow_empty'], bool):
        _allow_empty = kwargs['allow_empty']
        if _allow_empty:
            custom_regex = f"({custom_regex})?"  # Include the empty input in the regex
    else:
        _allow_empty = False
    if default is not None:
        _prompt += f" [default: {default}]"  # Inform the user about the default value
    # Questioning loop
    while True:
        inval = safe_string_input(prompt=_prompt, validations=custom_regex,
                                  error_msg="Error. Not a number, or expecting an integer but received a float",
                                  allow_empty=_allow_empty).replace(thousand_sep, '')
        if inval == '' and _allow_empty:  # If user input is empty, use the default value
            return default
        elif decimal_digits is None:
            try:
                if inval.find(decimal_sep) != -1:
                    inval = float(inval)
                else:
                    inval = int(inval)
            except ValueError:
                print(f"[ERROR] Unexpected numeric value ({inval})")
        elif decimal_digits > 0:
            inval = float(inval).__round__(decimal_digits)
        else:
            inval = int(inval)
        if (min_val is None or min_val <= inval) and (max_val is None or max_val >= inval):
            return inval
        else:
            print(f"[ERROR] Value out of bounds {bounds}")


def safe_list_input(max_size=-1, fixed_size=-1, sep=",", ltype="string", allow_repeated=True, validations=None,
                    prompt=None, confirm=False, allow_empty=True, **kwargs):
    """
    TODO: selectable list using the arrows
    :param max_size:
    :param fixed_size:
    :param sep:
    :param ltype:
    :param allow_repeated:
    :param validations:
    :param prompt:
    :param confirm:
    :param allow_empty:
    :param kwargs:
    :return:
    """
    _allowed_types = ['string', 'number']
    if fixed_size is not None and fixed_size > 0:
        print(f"[WARN] 'fixed_size' = {fixed_size} != {max_size} = 'max_size'."
              f" Overwriting 'max_size' value with 'fixed_size")
        max_size = fixed_size
    if prompt is None:
        pre_prompt = ''
    else:
        pre_prompt = f"{prompt}\n"
    while True:
        user_input = []
        if sep == 'new-line':
            # read each item in a new line
            # empty input to finish or max size reached
            # prompt = (i-th) value
            print(f"{pre_prompt}Press enter to add a value. Empty input to finish")
            counter = 1
            while True:
                if ltype == 'string':
                    inval = safe_string_input(
                        validations=validations,
                        prompt=f"({counter})"
                    )
                elif ltype == 'number':
                    inval = safe_number_input(
                        prompt=f"({counter})",
                        allow_empty=True,
                        **kwargs
                    )
                else:
                    raise ValueError("Unexpected value for parameter 'ltype'. Allowed: string|number")
                if inval is None or inval == '':
                    if fixed_size == -1 or len(user_input) == fixed_size:
                        break
                    else:
                        print(f"Required values: {fixed_size}. Missing: {fixed_size - len(user_input)}")
                        continue
                else:
                    if bool(allow_repeated) or inval not in user_input:
                        user_input.append(inval)
                        counter += 1
                    else:
                        print(f"[ERROR] Duplicates not allowed ({inval})")

                if max_size != -1 and len(user_input) >= max_size:
                    break
        else:
            # Single line input using sep
            # Read up to ``size`` items
            inval = safe_string_input(
                validations=validations,
                prompt=f"{pre_prompt}Enter the values separated by '{sep}':"
            )
            if max_size is not None and max_size > 0:
                _limit = max_size
            else:
                _limit = len(inval)
            if ltype == 'number':
                for num in [item.strip() for item in inval.split(sep)[0:_limit]]:
                    try:
                        if num.find('.') != -1:
                            user_input.append(float(num))
                        else:
                            user_input.append(int(num))
                    except ValueError:
                        print(f"[WARN] Invalid value found in the input list ({num})")
            elif ltype == 'string':
                if inval != '':
                    user_input.extend([item.strip() for item in inval.split(sep)[0:_limit]])
            else:
                raise ValueError("Unexpected value for parameter 'ltype'. Allowed: number|string")
            if not allow_repeated:
                user_input = list(set(user_input))
        if not allow_empty and len(user_input) == 0:
            print("Error. The list cannot be empty")
            continue
        elif fixed_size is not None and 0 < fixed_size != len(user_input):
            print(f"Error. Required values: {fixed_size}. Missing: {fixed_size - len(user_input)}")
            continue
        elif not confirm or (confirm and confirm_yes_no(user_input)):
            break
    return user_input


# def safe_number_list_input(size=-1, sep=",", min_val=0, max_val=0, decimal_digits=2, default=0, prompt=f"()",
#                            confirm=False):
#     # TODO: Differentiate between safe_number_list_input() and safe_string_list_input()?
#     pass

def selectable_list(args: list, placeholder=None, custom_prompt=None, enable_custom=False, verbose=False):
    # I disabled the param allow_empty because if a custom opt is offered, the empty option should not be available,
    # If you want to offer an "empty" option, you should include it in the list
    # :param allow_empty:
    #       Enables/disables empty input. Only applied on custom inputs (when ``enable_custom`` is ``True``)
    """
    TODO: Update selectable_list() docstring
    Examples:
        tcp_sc = selectable_list(scan_policy.tcp_scans(), placeholder=lambda x: f"Custom prefix {x}")
        recon_sc = selectable_list(scan_policy.recon_scans(), 'An additional scan')

    :param args: List of things (object, list, dictionaries) to be listed
    :param placeholder: function or lambda that generates a custom string for each element of the list.
    If it is None the string representation of the item is used. The strings are used for the list representation
    :param custom_prompt: custom function that request an input and return a value
    :param enable_custom: Enables/disables the custom option
    :param verbose: True to increase verbosity. False by default
    :return: A dict like: {'index': index of the selected element within the list, 'value': element selected}
    """
    if placeholder is None:
        placeholder = lambda x: x
    if (enable_custom is None or enable_custom is False) and custom_prompt is not None:
        # In the case of providing a custom prompt, enabled_custom is supposed to be True
        enable_custom = True
    aux = []  # List used to expand the selectable list with Cancel and Custom options
    # Generate the printable prompt
    indent = ' ' * 2
    counter = 0
    printable_prompt = ['Available options', '>', f'{indent}0. Cancel operation', '>']
    for elem in args:
        if isinstance(elem, list):
            for i in elem:
                counter += 1
                printable_prompt.insert(-1, f'{indent}{counter}. {placeholder(i)}')
                aux.append(i)
        # each dict item should not be selectable and independent of the rest of its sibling items
        elif isinstance(elem, dict):
            counter += 1
            val = text_padding(placeholder(f"\n{dict_to_table(elem)}"), padding=5)
            printable_prompt.insert(-1, f'{indent}{counter}. {val}')
        elif isinstance(elem, MenuEntry) and isinstance(elem.value, dict):
            counter += 1
            val = text_padding(placeholder(f"\n{dict_to_table(elem.value)}"), padding=5)
            printable_prompt.insert(-1, f'{indent}{counter}. {val}')
        else:
            counter += 1
            printable_prompt.insert(-1, f'{indent}{counter}. {placeholder(elem)}')
            aux.append(elem)
    # Add the custom option in the end. Maybe is better at the beginning?
    if enable_custom:
        counter += 1
        printable_prompt.insert(-1, f'{indent}{counter}. Custom value')
    while True:
        try:
            print("\n".join(printable_prompt))
            selection = int(input(f'\nPlease select one:\t'))
            if selection == 0:
                print(f'Operation canceled. Exiting...')
                exit(0)
            elif selection not in range(1, counter + 1):
                raise IndexError()
            else:
                if enable_custom and selection == counter:  # Custom option selected
                    # Counter = -1 => clarify that the (custom) option is not included in the list provided by the user
                    counter = -1
                    if custom_prompt is None:
                        value = safe_string_input(allow_empty=False)
                    else:
                        value = custom_prompt()
                else:  # Any other option selected
                    if verbose:
                        print(f"You selected: ({selection}) {aux[selection - 1]}")
                    counter = selection - 1  # Update counter to store the selected index
                    value = aux[selection - 1]
                break
        except ValueError:
            print(f'Error. Option not recognized')
        except IndexError:
            print(f'Error. Option out of bounds. Please select a number within the list')
    return {"index": counter, "value": value}


# --------------------------------------------------------------------------
# --------------------<| M E N U |>-----------------------------------------
# --------------------------------------------------------------------------
class MenuEntry:
    def __init__(self, val, desc=None, callback=None):
        self.value = val
        if desc is not None:
            self.description = str(desc)
        else:
            self.description = None
        self.callback = callback

    def __str__(self):
        if self.description is not None and self.description != self.value:
            return f"'{str(self.value)}' ({self.description})"
        else:
            return f"'{str(self.value)}'"

    def __repr__(self):
        return str(self.value)


class Menu:
    def __init__(self, title=None, desc=None, allowcustom=False, customprompt=None):
        if title is None:
            self.title = 'Untitled menu'
        else:
            self.title = f"{title} menu"
        if desc is None:
            self.description = 'no description provided'
        else:
            self.description = str(desc)
        self.entries = []
        self.allow_custom = allowcustom
        self.custom_prompt = customprompt

    def add_entry(self, value, description=None, callback=None):
        self.entries.append(MenuEntry(value, description, callback))

    def add_entries(self, iterable: Union[dict, list]):
        """

        :param iterable: a dict like {'option1 description': 'option1 value', opt2...}
            or a list of serialized (dictionaries) MenuEntry instances
        :return:
        """
        if isinstance(iterable, dict):
            for k, v in iterable.items():
                self.add_entry(v, k)
        elif isinstance(iterable, list):
            for elem in iterable:
                if isinstance(elem, dict) and 'value' in elem:
                    val = elem['value']
                else:
                    val = elem
                if isinstance(elem, dict) and 'description' in elem:
                    desc = elem['description']
                else:
                    desc = None
                if isinstance(elem, dict) and 'callback' in elem:
                    cb = elem['callback']
                else:
                    cb = None
                self.add_entry(val, description=desc, callback=cb)
        else:
            raise TypeError("Unexpected type for parameter 'iterable'. Allowed: dict | list")

    def open(self, verbose=False):
        if verbose:
            print(f"{self.title} - {self.description}")
        ret = selectable_list(self.entries, enable_custom=self.allow_custom, custom_prompt=self.custom_prompt)
        if isinstance(ret['value'], MenuEntry) and ret['value'].callback is not None:
            ret['value'].callback()
        return ret


# --------------------------------------------------------------------------
# --------------------<| F O R M S |>---------------------------------------
# --------------------------------------------------------------------------

class StringField:
    def __init__(self, name: str, description='', example='', ftype='string', regex=None, allow_empty=True,
                 confirm=False, defaultval=None):
        _allowed_types = ['string', 'numeric', 'list']
        if ftype not in _allowed_types:
            raise ValueError(f"Invalid value for FormField.ftype. Allowed values are: {', '.join(_allowed_types)}")
        elif not isinstance(allow_empty, bool):
            raise TypeError("Invalid value for FormField.is_nullable. Only bool allowed")
        else:
            self.ftype = str(ftype)
            self.name = str(name)
            self.description = str(description)
            self.example = str(example)
            self.confirm = bool(confirm)
            self.regex = regex
            self.allow_empty = bool(allow_empty)
            self.default_value = defaultval

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        name = self.name
        desc = self.description
        eg = self.example
        if self.description == '':
            desc = 'Description not provided'
        if self.example == '':
            eg = 'not provided'
        return f"'{name}': {desc}. (e.g.: {eg})"

    def long_str(self):
        name = self.name
        desc = self.description
        eg = self.example
        if self.description == '':
            desc = 'Not provided'
        if self.example == '':
            eg = 'Not provided'
        return f"  Title: {name}\n  Description: {desc}\n  Example: {eg}"

    def input(self):
        # while True:
        return safe_string_input(
            validations=self.regex,
            prompt=f'>> {self.name} = ',
            confirm=self.confirm,
            allow_empty=self.allow_empty
        )
        # inval = input(f'>> {self.name} = ')
        # if inval == '' and not self.is_nullable:
        #     print(f'[ERROR] This field cannot be empty and must match')
        #     continue
        # # Check field syntax
        # if self.regex is not None and re.fullmatch(self.regex, inval) is None:
        #     print(f'[ERROR] Invalid input syntax')
        #     continue
        # return inval


class NumericField(StringField):
    def __init__(self, name, decimal_digits=None, decimal_separator='.', min_val=None, max_val=None, **kwargs):
        _allowed_separators = [',', '.']
        if min_val > max_val:
            raise ArithmeticError("Parameter 'min_val' is greater than 'max_val'")
        elif decimal_separator not in _allowed_separators:
            raise ValueError("Unexpected value for the parameter 'decimal_separator'. Allowed: ',' | '.'")
        else:
            super().__init__(name=name, ftype='numeric', **kwargs)
            self.min_val = min_val
            self.max_val = max_val
            self.decimal_digits = decimal_digits
            self.decimal_separator = decimal_separator

    @property
    def get_regex(self):
        # TODO: Build regex from min_val, max_val, decimal_digits and decimal_separator
        custom_regex = "^"
        if self.min_val < 0 or self.max_val < 0:
            custom_regex += "-?"
        if self.max_val is None:
            custom_regex += "[0-9]+"
        else:
            custom_regex += "[0-9]{1," + str(
                len(str(self.max_val).replace('-', '').split(self.decimal_separator))) + "}"
        if self.decimal_digits > 0:
            custom_regex += f"(\\{self.decimal_separator}[0-9]" + "{1," + str(self.decimal_digits) + "})?"
        return custom_regex

    def input(self):
        return safe_number_input(
            min_val=self.min_val,
            max_val=self.max_val,
            decimal_digits=self.decimal_digits,
            decimal_sep=self.decimal_separator,
            default=self.default_value,
            prompt=f'>> {self.name} = ',
            confirm=self.confirm,
            allow_empty=self.allow_empty
        )
        # if self.decimal_digits > 0:
        #     inval = float(inval)
        # else:
        #     inval = int(inval)
        # if self.min_val <= inval <= self.max_val:
        #     return inval
        # else:
        #     raise ValueError(f"Value out of range. Allowed: [{self.min_val}, {self.max_val}]")


class ListField(StringField):
    def __init__(self, name, sep=',', max_size=-1, fixed_size=-1, ltype="string", allow_repeated=True, **kwargs):
        if ltype in ['string', 'number']:
            super().__init__(name=name, ftype='list', **kwargs)
            self.ltype = ltype
            self.sep = sep
            self.max_size = max_size
            self.fixed_size = fixed_size
            self.allow_repeated = bool(allow_repeated)
            self.properties = kwargs
        else:
            raise ValueError("Unexpected value for parameter 'ltype'. Allowed: string|number")

    def input(self):
        return safe_list_input(
            max_size=self.max_size,
            fixed_size=self.fixed_size,
            sep=self.sep,
            ltype=self.ltype,
            allow_repeated=self.allow_repeated,
            validations=self.regex,
            confirm=self.confirm,
            allow_empty=self.allow_empty,
            **self.properties
        )


class PromptForm:
    """
    Contrary to the class ``Menu``, which allows the user to select an option from a list,
    the Form class allows the user to input values
    """

    def __init__(self, title='untitled', desc=''):
        self.title = str(title)
        self.description = str(desc)
        self.fields = []
        self.last_result = {}
        # TODO: usar pandas
        # las columnas son los tries
        # Las filas los campos del formulario

    def __str__(self):
        return f"PromptForm('{self.title}')"

    def show_info(self):
        print(f"---< Form '{self.title}' >---\n{self.description}\n---------")
        for f in self.fields:
            print(f"- {f}")

    def get_summary_table(self):
        return f"\n----< Form '{self.title}' >----\n{dict_to_table(self.last_result)}"

    def add_string_field(self, name, **kwargs):
        self.fields.append(StringField(name=name, **kwargs))

    def add_numeric_field(self, name, **kwargs):
        self.fields.append(NumericField(name=name, **kwargs))

    def add_list_field(self, name, **kwargs):
        self.fields.append(ListField(name=name, **kwargs))

    def reset(self):
        self.last_result.clear()

    def fill_in(self, compact_mode=True, confirm=False):
        while True:
            input_form = {}
            print(f"---< Form '{self.title}' >---")
            if not compact_mode:
                print(f"{self.description}\n{'-' * 20}")
            field_num = 0
            while True:
                if field_num >= len(self.fields):
                    break
                field = self.fields[field_num]
                if isinstance(field, StringField):
                    # Print field information
                    prompt = f"[Field {field_num + 1}/{len(self.fields)}]"
                    if compact_mode:
                        prompt += f" {field}"
                    else:
                        prompt += f"\n{field.long_str()}"
                    print(prompt)
                    input_form[field.name] = field.input()
                    field_num += 1
                else:
                    raise TypeError("Invalid field type. Only cli.StringField allowed")
            self.last_result.update(input_form)
            if not confirm or (confirm and confirm_yes_no(self.get_summary_table())):
                break
            else:
                continue
        return input_form

    def export(self):
        # TODO: cli.PromptForm.export()
        # Exports the lasts results into a txt or json file
        pass


# --------------------------------------------------------------------------
# --------------------<| M A I N |>-----------------------------------------
# --------------------------------------------------------------------------


if __name__ == '__main__':
    # Do some quick tests here
    pass



import webbrowser

# def selenium_scrapper():
#     # selenium 4
#     from selenium import webdriver
#     from selenium.webdriver.common.by import By
#     from selenium.webdriver.firefox.service import Service as FirefoxService
#     from webdriver_manager.firefox import GeckoDriverManager
#
#     driver = webdriver.Firefox(service=FirefoxService(GeckoDriverManager().install()))
#     driver.get(f'https://google.com')
#     title = driver.title
#     print(title)
#     links = driver.find_elements(By.TAG_NAME, "a")
#     for ln in links:
#         print(ln.get_property('href'))
#     print(f'{len(links)} links found')
#     driver.quit()


def open_browser(self):
    # TODO: check installation. Offer installation
    browsers = {
        "firefox": "C:\\Program Files\\Mozilla Firefox\\firefox.exe",
        "edge": "C:\\Program Files (x86)\\Microsoft\\Edge\\Application\\msedge.exe",
        "chrome": "C:\\Program Files\\Google\\Chrome\\Application\\chrome.exe",
        "chromium": "C:\\Program Files (x86)\\chromium-win\\chrome-win\\chrome.exe",
        "brave": "C:\\Program Files\\BraveSoftware\\Brave-Browser\\Application\\brave.exe",
        "opera": "C:\\Users\\rhleon\\AppData\\Local\\Programs\\Opera\\launcher.exe"
    }
    # To register
    for br_name, br_path in browsers.items():
        webbrowser.register(br_name, None, webbrowser.BackgroundBrowser(br_path))

    # To open
    for browser in browsers:
        webbrowser.get(browser).open('http://www.google.com')



'''This module implements specialized container datatypes providing
alternatives to Python's general purpose built-in containers, dict,
list, set, and tuple.

* namedtuple   factory function for creating tuple subclasses with named fields
* deque        list-like container with fast appends and pops on either end
* ChainMap     dict-like class for creating a single view of multiple mappings
* Counter      dict subclass for counting hashable objects
* OrderedDict  dict subclass that remembers the order entries were added
* defaultdict  dict subclass that calls a factory function to supply missing values
* UserDict     wrapper around dictionary objects for easier dict subclassing
* UserList     wrapper around list objects for easier list subclassing
* UserString   wrapper around string objects for easier string subclassing

'''
"""
__version__ = '0.0.3'
__all__ = [
    'workers'
    'ChainMap',
    'Counter',
    'OrderedDict',
    'UserDict',
    'UserList',
    'UserString',
    'defaultdict',
    'deque',
    'namedtuple',
]
__author__ = 'Ramón Hernández León <rhdezl05@gmail.com>'

import workers as workers
import cli as cli
import gui as gui
import utilities as utilities

try:
    from _collections import deque
except ImportError:
    pass
else:
    _collections_abc.MutableSequence.register(deque)

def ChainMap():
"""


import klsframe.cli.cli as cli


def test_prompt_form_simple():
    form = cli.PromptForm(title='Test 1', desc='Just a test. Or not. I don\'t care')
    form.add_field('nombre', nullable=False)
    form.add_field('apellidos', nullable=False)
    res = form.fill_in()
    print(res)
    assert res == form.last_result
    assert 'nombre' in res and 'apellidos' in res
    print('[OK] Test prompt_form_simple_test succeeded\n\n')


def test_prompt_form_advanced():
    form = cli.PromptForm(title='Test 2', desc='Just a test. Or not. I don\'t care')
    form.add_string_field(name='Nombre', description='Nombre del titular de la cuenta', example='Pedro')
    form.add_string_field(name='Apellidos', description='Apellidos del titular de la cuenta', example='Pérez')
    form.add_numeric_field(name='Edad', description='Edad en años del titular de la cuenta', example='12',
                           decimal_digits=0, min_val=1, max_val=99)
    form.add_list_field(name='hobbies', description="Top-3 hobbies", max_size=3)
    form.add_list_field(name='fav num', description='Top-3 Favorite numbers', ltype='number', allow_repeated=False,
                        fixed_size=3)
    # form.show_info()
    res = form.fill_in(compact_mode=True, confirm=True)
    print(res)
    assert res == form.last_result
    print('[OK] Test prompt_form_advanced_test succeeded\n\n')


def test_confirm():
    print(cli.confirm_yes_no("Delete all assets", default=False, allow_empty=False, shortened=False))
    print(cli.confirm_yes_no("Exit", default=True, allow_empty=True, shortened=True))
    print(cli.confirm_yes_no("Eres tonoto?"))


def test_safe_input():
    age1 = input("Age")
    age2 = cli.safe_string_input("Age", )


def test_menu_1():
    def cagar():
        print('prffrfffrf')

    def comer():
        print('ñamñam')

    def dormir():
        print('zZzzZZZzZzzZZz')

    men = cli.Menu(allowcustom=True)
    men.add_entry('cagar', 'Ir al trono', cagar)
    men.add_entry('comer', 'Monchear', comer)
    men.add_entry('dormir', 'Sobarla', dormir)
    print(men.open())


def test_menu_2():
    dict_port_listing = {"Por defecto (no se especifican puertos)": "",
                         "Todos los puertos": "-p-",
                         "1000 puertos comunes": "--top-ports 1000"}
    menu1 = cli.Menu(allowcustom=True)
    menu1.add_entries(dict_port_listing)
    print(menu1.open())


def test_menu_3():
    menu_options1 = [
        {'value': '', 'description': 'Por defecto (no se especifican puertos)'},
        {'value': '-p-', 'description': 'Todos los puertos'},
        {'value': '--top-ports 1000', 'description': '1000 puertos comunes'}
    ]

    def input_port():
        prompt = f"Introduce tu propia seleccion de puertos, separados por espacio (23 80 443) o un rango (0-1024)"
        range_regex = '^[0-9]{1,5}-[0-9]{1,5}$'
        listing_regex = '^(([0-9]{1,5})+[ ]?)+$'
        return cli.safe_string_input(validations=[range_regex, listing_regex], allow_empty=False, prompt=prompt)

    menu2 = cli.Menu(allowcustom=True, customprompt=input_port)
    menu2.add_entries(menu_options1)
    print(menu2.open())


def test_menu_4():
    data = ["Por defecto (no se especifican puertos)", "Todos los puertos", "1000 puertos comunes"]
    menu1 = cli.Menu(title='Test 4', desc='Just a dummy test. [Implicit options: (-T3, -vvv)]', allowcustom=True)
    menu1.add_entries(data)
    print(menu1.open(verbose=True))


def test_menu_5():
    data = [{'tipo': 'tcp', 'desc': 'TCP SYN scan top ports', 'params': '-Pn -n -sS --stats-every 15s'},
            {'tipo': 'tcp', 'desc': 'TCP SYN scan ALL ports', 'params': '-p- -Pn -n -sS --stats-every 15s'}]
    menu1 = cli.Menu(title='Test 4', desc='Just a dummy test. [Implicit options: (-T3, -vvv)]', allowcustom=True)
    menu1.add_entries(data)
    print(menu1.open(verbose=True))


def test_safe_list_input_1():
    print(cli.safe_list_input(sep='new-line', max_size=5, fixed_size=3, ltype='number', allow_repeated=False,
                              prompt='Dime 3 numeros del 1 al 10', min_val=1, max_val=10))


def test_continue_or_exit():
    print("hola?")
    cli.continue_or_exit()
    print("adios!")

if __name__ == '__main__':
    # test_confirm()
    # test_prompt_form_simple()
    # test_prompt_form_advanced()
    # test_dict_to_table()
    # pyautogui.prompt(text="Apellido", title="field 1/3", default="mondongo")
    print(test_menu_5())



import collections
import json
import os
import re

import pyautogui
# import klsframe.workers.desktopWorker as worker

Point = collections.namedtuple("Point", "x y")
Size = collections.namedtuple("Size", "width height")

"""
;button;input;textbox;hyperlink;image;table
inner text;?;Y;Y;Y;N;Y
"""


def calibrate():
    # 1. Select: new gui layout OR edit existing one
    # 2. Select: view current layout OR edit current layout
    #              (tree view)            (list pages)
    # 3. Add a page OR edit an existing page
    # 4. Add / Edit / Remove component
    pass


def safe_coordinates(x, y):
    if x is None or y is None:
        return False
    else:
        return not (x <= 0 or pyautogui.size().width <= x or y <= 0 or pyautogui.size().height <= y)


def screen_center():
    return pyautogui.size().width / 2, pyautogui.size().height / 2


class GUILayout:
    """
    A GUI Layout is a list of pages.
    """

    def __init__(self):
        self.name = ''
        self.screen_size = pyautogui.size()
        self.pages = []

    def __eq__(self, other):
        if other is None or len(self.pages) != len(other.pages):
            return False
        for pg_i, pg_j in zip(self.pages, other.pages):
            if pg_i.name != pg_j.name:
                return False
            if pg_i.components != pg_j.components:
                return False
        return True

    def get(self, comp_name, pg_name=None):
        for pg in self.pages:
            if pg_name is not None and pg_name != pg.name:
                continue
            for k, v in pg.components.items():
                if k == comp_name:
                    return v
        return None

    def new_page(self, page_name):
        _new = GUIPage(page_name)
        self.pages.append(_new)
        return _new

    def add_component(self, page_name, comp_name, x, y):
        for pg in self.pages:
            if pg.name == page_name:
                pg.components[comp_name] = (x, y)
                return
        _new_pg = self.new_page(page_name)
        _new_pg.components[comp_name] = (x, y)
        return

    def to_json(self):
        _res = []
        counter = -1
        for pg in self.pages:
            counter += 1
            _res.append({pg.name: {}})
            for comp, comp_coors in pg.components.items():
                _res[counter][pg.name][comp] = [comp_coors[0], comp_coors[1]]
        return _res

    @staticmethod
    def load(source):
        _gly = GUILayout()
        # TODO: load from a json file/object
        if type(source) is list:
            for src in source:
                _new_pg = _gly.new_page("".join(src.keys()))
                for comp_name, comp_coords in src[_new_pg.name].items():
                    _gly.add_component(_new_pg.name, comp_name, x=comp_coords[0], y=comp_coords[1])
        elif type(source) is str and os.path.isfile(source):
            with open(source, 'r') as file:
                return GUILayout.load(json.load(file))
        elif type(source) is str:
            # TODO: implement JSON string parsing
            pass
        else:
            raise ValueError("Invalid object type for load()")
        return _gly

    def local_save(self, outfile):
        with open(outfile, 'w') as file:
            json.dump(self.to_json(), file, indent=2)

    def verify(self):
        for pg in self.pages:
            for comp, comp_coords in pg.components.items():
                if not safe_coordinates(comp_coords[0], comp_coords[0]):
                    print(f"ERROR. Coordinate verification failed. Coordinates of component '{comp}' are not valid")
                    return 1
        return 0

    # START TABLE FUNCTIONS
    def get_table_row_at(self, tname, rownum):
        pass

    def get_table_row_with(self, tname, filters):
        pass

    def get_table_all_rows(self, tname):
        pass

    # END TABLE FUNCTIONS

    def screen_center(self):
        return self.screen_size.width / 2, self.screen_size.height / 2


class GUIPage:
    def __init__(self, pgname):
        self.name = str(pgname)
        self.transitions = []
        self.components = []
        self.tables = []

    def __repr__(self):
        return self.__dict__


class GUITransition:
    # TODO
    pass


class GUIComponent:
    """
    comp_name: "boton borrar"
    comp_type: "button"
    start: (x1, y1)
    end: (x2, y2)
    screenArea
    regex: "borrar"
    """

    def __init__(self, title, compclass: str, regex=None):
        _allowed_classes = ['button', 'input', 'textbox', 'table', 'hyperlink', 'image']
        # Check component type
        if str(compclass.lower()) in _allowed_classes:
            self.comp_type = str(compclass.lower())
            self.title = title
            self.regex = regex
            # self.start = None
            # self.end = None
            # self.screen_area = None
            self.screen_area = ScreenArea()
        else:
            raise TypeError(f"Invalid type ({type}) for GUIComponent")

    def calibrate(self):
        self.screen_area = ScreenArea()

    def get_inner_text(self):
        return self.title

    def validate(self, value):
        if self.regex is None or re.fullmatch(self.regex, value) is not None:
            return True
        else:
            return False

    def get_point(self, position='center'):
        _allowed_positions = ['top-left-corner', 'top-right-corner', 'bottom-right-corner', 'bottom-left-corner',
                              'left-side', 'right-side', 'top-side', 'bottom-side',
                              'left-border', 'top-border', 'right-border', 'bottom-border',
                              'top-left-side', 'top-right-side', 'bottom-right-side', 'bottom-left-side', 'center']
        if position not in _allowed_positions:
            raise ValueError(f"Unexpected value for parameter 'position'. Allowed: {', '.join(_allowed_positions)}")
        return eval(f"self.screen_area.{opt.replace('-', '_')}")
        # w = worker.Worker()
        # w.click_at(eval(f"self.screen_area.{opt.replace('-', '_')}"), button=button, clicks=clicks)


class GUIButton(GUIComponent):
    def __init__(self, title, regex=None):
        super().__init__(compclass='button', title=title, regex=regex)


class GUIInput(GUIComponent):
    def __init__(self, title, regex=None):
        super().__init__(compclass='input', title=title, regex=regex)


class GUITextBox(GUIComponent):
    def __init__(self, title, regex=None):
        super().__init__(compclass='textbox', title=title, regex=regex)


class GUIHyperLink(GUIComponent):
    def __init__(self, title, regex=None):
        super().__init__(compclass='hyperlink', title=title, regex=regex)


class GUIImageBox(GUIComponent):
    def __init__(self, title, regex=None):
        super().__init__(compclass='image', title=title, regex=regex)


class GUITable(GUIComponent):
    def __init__(self, title):
        super().__init__(title=title, compclass='table')
        self.first_row_Y = 0.0
        self.row_height = 0.0
        self.table_size = 0  # Num of rows
        self.next_page_button = GUIButton(title='nextPage')
        self.prev_page_button = GUIButton(title='prevPage')
        self.columns = []

    def get_row_at(self, index):
        # TODO: Fix this shit. GuiLayour CANNOT use WORKER.
        # IT only hold coordinates values, but it cannot click, nor move
        pass
        # if index == 0:
        #     return self.headers()
        # elif 0 < index <= self.table_size:
        #     # TODO
        #     offset = (index - 1) * self.row_height
        #     row = {}
        #     for col in self.columns:
        #         # A human will just read the columns, but the app need to copy the text to get it
        #         # Thus those clicks do not count for Worker.stats
        #         w = worker.Worker()
        #         content = w.copy_text(x=col.left_border(), y=self.first_row_Y + offset)
        #         if self.validate(content):
        #             row[col.col_name] = content
        #         else:
        #             raise ValueError(f"Unexpected value ({content}) for column '{col.col_name}'")
        #     return row
        # else:
        #     raise IndexError(f"Index {index} out of bounds [0, {self.table_size}]")

    def get_row_where(self, **kwargs):
        pass

    def headers(self):
        return [col.col_name for col in self.columns]


class GUITableColumn:
    def __init__(self):
        self.col_name = ''
        self.left_border_x = -1.0
        self.right_border_x = -1.0
        self.regex = None


class ScreenArea:
    def __init__(self, startp=None, endp=None):
        self.start = None
        self.end = None
        # Check coordinates
        if startp is not None and endp is not None:
            self.set_coordinates('start', startp)
            self.set_coordinates('end', endp)
        elif (startp is None) ^ (endp is None):
            raise AttributeError("The parameter 'startp' and 'endp' must be used together")

    def set_coordinates(self, position, point):
        # Check coordinates
        if not isinstance(point, (Point, tuple)):
            raise TypeError("Invalid type for parameter 'point'. Must be coordinates tuples (x, y)")
        if position == 'start':
            if self.end is None:
                self.start = point
            elif point[0] >= self.end.x or point[1] >= self.end.y or point <= (0, 0):
                raise ValueError("Invalid coordinates. Start is bigger than end")
            else:
                self.start = point
        elif position == 'end':
            if self.start is None:
                self.end = point
            elif point[0] <= self.start.x or point[1] <= self.start.y or point <= (0, 0):
                raise ValueError("Invalid coordinates. End is smaller than start")
            else:
                self.end = point
        else:
            raise ValueError(f"Unknown position '{position}'. Valid: start|end")

    @staticmethod
    def select_screen_area(verbose=False):
        import klsframe.IplusD.loggers as logger
        while True:
            try:
                print(f"[INFO] Select a screen area. Click the top-left border, then the bottom-right one.")
                user_input = logger.record_mouse(clicks=2, verbose=verbose)
                sa = ScreenArea()
                sa.set_coordinates('start', user_input[0]['position'])
                sa.set_coordinates('end', user_input[1]['position'])
                return sa
            except (ValueError, TypeError) as e:
                sa = None
                print(e)

    # ----| C E N T E R |----
    @property
    def center(self) -> Point:
        return Point(self.center_x, self.center_y)

    @property
    def center_x(self) -> int:
        return self.start.x + self.width / 2

    @property
    def center_y(self) -> int:
        return self.start.y + self.height / 2

    @property
    def width(self) -> int:
        return self.end.x - self.start.x

    @property
    def height(self) -> int:
        return self.end.y - self.start.y

    @property
    def size(self) -> Size:
        return Size(self.width, self.height)

    # ----| S I D E S |----
    @property
    def top_left_side(self):
        return Point(self.start.x + self.width * 0.25, self.start.y + self.height * 0.25)

    @property
    def top_right_side(self):
        return Point(self.end.x - self.width * 0.25, self.start.y + self.height * 0.25)

    @property
    def bottom_left_side(self):
        return Point(self.start.x + self.width * 0.25, self.end.y - self.height * 0.25)

    @property
    def bottom_right_side(self):
        return Point(self.end.x - self.width * 0.25, self.end.y - self.height * 0.25)

    @property
    def left_side(self):
        return Point(self.start.x + self.width * 0.25, self.center_y)

    @property
    def right_side(self):
        return Point(self.end.x - self.width * 0.25, self.center_y)

    @property
    def top_side(self):
        return Point(self.center_x, self.start.y + self.height * 0.25)

    @property
    def bottom_side(self):
        return Point(self.center_x, self.end.y - self.height * 0.25)

    # ----| B O R D E R S |----
    def left_border(self):
        return Point(self.start.x, self.center_y)

    def right_border(self):
        return Point(self.end.x, self.center_y)

    def top_border(self):
        return Point(self.center_x, self.start.y)

    def bottom_border(self):
        return Point(self.center_x, self.end.y)

    # ----| C O R N E R S |----
    def corners(self, reverse=False) -> tuple:
        """

        :return: a 4-position tuple (top-left, top-right, bottom-right, bottom-left)
                indicating the corner points of the area
        """
        if isinstance(reverse, bool) and reverse:
            return (self.top_left_corner, self.bottom_left_corner,
                    self.bottom_right_corner, self.top_right_corner)

        elif isinstance(reverse, bool):
            return (self.top_left_corner, self.top_right_corner,
                    self.bottom_right_corner, self.bottom_left_corner)
        else:
            raise TypeError("Invalid type for parameter 'reverse'. Expected: bool")

    @property
    def top_left_corner(self) -> Point:
        return Point(self.start.x, self.start.y)

    @property
    def top_right_corner(self) -> Point:
        return Point(self.end.x, self.start.y)

    @property
    def bottom_left_corner(self) -> Point:
        return Point(self.start.x, self.end.y)

    @property
    def bottom_right_corner(self) -> Point:
        return Point(self.end.x, self.end.y)

    def draw_border(self, rounds=1, drawtime=5.0, reverse=False):
        tmp = pyautogui.PAUSE  # Disable pause to meet the specified drawtime
        pyautogui.PAUSE = 0
        if drawtime > 1.0:
            _duration = drawtime / 5.0
        else:
            raise ValueError("The drawing time must be greater than 0")
        _rectangle = self.corners(reverse=reverse)
        for i in range(0, rounds):
            for j in range(0, 5):
                point = _rectangle[j % 4]
                # TODO: implement the moves with kls-worker
                pyautogui.moveTo(x=point[0], y=point[1], duration=_duration)
        pyautogui.PAUSE = tmp

    def select(self):
        # TODO: develop select (hold ctrl + drag)
        pass


if __name__ == '__main__':
    sa1 = ScreenArea.select_screen_area(verbose=True)
    # sa1.draw_border(rounds=2, reverse=True, drawtime=0.5)
    opt = 'top-left-corner'
    print(eval(f"self.{opt.replace('-', '_')}"))



import re


def equals_ignore_case(str1, str2):
    return re.fullmatch(str1, str2, flags=re.IGNORECASE) is not None


def validate(value, validations):
    if validations is None:
        raise ValueError("Validations cannot be None")
    elif not isinstance(validations, list):
        validations = [validations]
    if len(validations) == 0:
        return True
    else:
        for valid in validations:
            if re.fullmatch(valid, value):
                return True
        return False


def dict_to_table(dictionary, compact=False):
    if not isinstance(dictionary, dict):
        raise TypeError("Invalid type for parameter 'dictionary'. Must be a Python dictionary")
    longest_key = -1
    longest_value = -1
    for k, v in dictionary.items():
        if len(str(k)) > longest_key:
            longest_key = len(str(k))
        if len(str(v)) > longest_value:
            longest_value = len(str(v))

    if not isinstance(compact, bool):
        raise TypeError("Invalid type for the parameter 'compact'. It must be bool")
    elif compact:
        pad = "_"
        table_border = f"{'_' * (longest_key * 2 + 8)}"
        final = [table_border, "\n"]
    else:
        pad = " "
        table_border = f"+{'-' * (longest_key + 2)}+{'-' * (longest_value + 2)}+"
        final = [table_border, table_border + "\n"]
    for k, v in dictionary.items():
        final.insert(
            len(final) - 1,
            f"|{pad}{k}{pad * (longest_key - len(str(k)))}{pad}|{pad}{v}{pad * (longest_value - len(str(v)))}{pad}|"
        )
    return "\n".join(final)


def text_padding(msg, padding=2, decorator=None, decorate_lines=None):
    valid_dl = [None, 'all', 'first', 'last', 'even', 'odd', []]
    if (decorator is None) ^ (decorate_lines is None):
        raise ValueError("The parameters 'decorator' and 'decorate_lines' must be used together")
    if not isinstance(padding, int):
        raise ValueError("The parameter 'padding' must be a integer number")
    counter = 0
    final = []
    for ln in msg.split('\n'):
        counter += 1
        if decorator is None:
            final.append(f'{" " * padding}{ln}')
        else:
            if (decorate_lines is None) or decorate_lines == '' or \
                    (isinstance(decorate_lines, list) and counter not in decorate_lines) or \
                    (decorate_lines == 'first' and counter != 1) or \
                    (decorate_lines == 'last' and counter != len(msg.split('\n'))) or \
                    (decorate_lines == 'even' and counter % 2 != 0) or \
                    (decorate_lines == 'odd' and counter % 2 == 0):
                _decorator = ' '
            elif not isinstance(decorate_lines, list) and decorate_lines not in valid_dl:
                raise ValueError("Unexpected value for parameter 'decorate_lines'")
            else:
                _decorator = decorator
            final.append(f'{" " * padding}{_decorator} {ln}')
    return "\n".join(final)


def sort_dict(dictionary: dict, sortby='keys', reverse=False):
    if sortby == 'keys':
        sort_key = lambda ele: ele[0]
    elif sortby == 'values':
        sort_key = lambda ele: ele[1]
    else:
        raise ValueError("Invalid value for parameter 'sortby'. Valid: keys | values")
    return {key: val for key, val in sorted(dictionary.items(), key=sort_key, reverse=reverse)}


def join_by(join_char, iterable):
    """
    Joins an <iterable> with <join_char>. Uses list-comprehension
    :param join_char: Char to be used as join
    :param iterable: Iterable to be joined using <join_char>
    :return: A string with all the elements of <iterable> joined by <join_char>
    """
    return str(join_char).join([str(val) for val in iterable])


if __name__ == '__main__':
    # Do some quick tests here
    pass



import sys
import time
import random
import re
import signal

# Problematic imports - start
import keyboard
import pyautogui
import pyperclip
# Problematic imports - end

import klsframe.gui.guiLayout as guiLayout


# --------------------------------------------------------------------------
# BASIC API CALLS
# --------------------------------------------------------------------------

def click(x=None, y=None, button='left', clicks=1, interval=0.0):
    """
    Moves to (x, y) and clicks the specified button ``clicks`` times
    :param x:
    :param y:
    :param button:
    :param clicks:
    :param interval:
    :return:
    """
    pyautogui.click(x=x, y=y, button=button, clicks=clicks, interval=interval)


def click_on(elem: guiLayout.GUIComponent, button='left', clicks=1):
    pt = elem.get_point(position='center')
    click_at(pt, button=button, clicks=clicks)


def click_at(point: guiLayout.Point, button='left', clicks=1):
    click(x=point.x, y=point.y, button=button, clicks=clicks)


def move(x, y, duration=0.0):
    pyautogui.moveTo(x, y, duration=duration, tween=pyautogui.easeOutQuad)


def move_to(self, point: guiLayout.Point):
    pyautogui.moveTo(point.x, point.y, duration=self.speed_profile, tween=pyautogui.easeOutQuad)


def write(msg, interval=0.0):
    """
    Tries to write wherever the cursor is placed

    Arguments
        msg. String to be written. Can contain special characters (\n, \t...)
        If msg is a list of strings, then each element of the list will be written in a new line
    """
    if type(msg) is str:
        pyautogui.write(message=msg, interval=interval)
    elif type(msg) is list:
        for m in msg:
            pyautogui.write(message=f'{m}\n', interval=interval)
    else:
        raise TypeError('Invalid type for msg in write()')


def press(*args, interval=0.0):
    """
    Combination of pyautogui.press and pyautogui.hotkey
    press('ctrl', 'c') = press(*['ctrl', 'c']) = press('ctrl+c')
    You can pass a list of key-actions that will be executed in order, for example: 'ctrl+c', 'ctrl+v'
    Arguments (2 types)
        string
             'a' -> single press
             'a+b+c' -> hotkey (use \\ to escape +, like 3\\+2 = 5)
        tuple
            ('c', 2) -> {'key': 'c', 'presses': 2}
            ('ctrl+t', 2) -> {'key': 'ctrl+t', 'presses': 2}
    """
    for arg in args:
        if type(arg) is tuple:
            _key = arg[0]
            _presses = arg[1]
        elif type(arg) is str:
            _key = arg
            _presses = 1
        else:
            raise TypeError('Invalid args type in press() ')

        if _key.find('+') != -1 and re.search('^\\w+(\\+\\w+)+', arg) is not None:
            for _i in range(0, _presses):
                pyautogui.hotkey(*_key.split('+'))
        else:
            pyautogui.press(keys=_key, presses=_presses, interval=interval)


def scroll(value, direction='vertical', x=None, y=None):
    """
    pyautogui.scroll(10)   # scroll up 10 "clicks"
    pyautogui.scroll(-10)  # scroll down 10 "clicks"
    pyautogui.scroll(10, x=100, y=100)  # move mouse cursor to 100, 200, then scroll up 10 "clicks"

    ONLY in OS X and Linux platforms
    pyautogui.hscroll(10)   # scroll right 10 "clicks"
    pyautogui.hscroll(-10)   # scroll left 10 "clicks"
    """
    _allowed_directions = ['vertical', 'horizontal']
    _allowed_values = range(-10000, 10001)
    if direction not in _allowed_directions:
        raise ValueError(f"Error. Invalid value for direction ({direction})")
    elif direction == 'horizontal' and sys.platform.find('win') != -1:
        raise ValueError(f"Error. Horizontal scrolling not supported on Windows platform")
    elif value not in _allowed_values:
        raise ValueError(f"Error. Invalid scroll value. Valid range [-100, 100]")

    if direction == 'vertical':
        vt_scroll(clicks=value, x=x, y=y)
    elif direction == 'horizontal':
        hz_scroll(clicks=value, x=y, y=y)
    else:
        raise ValueError('Unexpected value for argument "direction"')


def hz_scroll(clicks, x=None, y=None):
    if x is not None and y is not None:
        if guiLayout.safe_coordinates(x, y):
            pyautogui.hscroll(clicks=clicks, x=x, y=y)
        else:
            raise ValueError(f"Error. Coordinates ({x}, {y}) out of range")
    else:
        pyautogui.hscroll(clicks=clicks)


def vt_scroll(clicks, x=None, y=None):
    if x is not None and y is not None:
        if guiLayout.safe_coordinates(x, y):
            pyautogui.vscroll(clicks=clicks, x=x, y=y)
        else:
            raise ValueError(f"Error. Coordinates ({x}, {y}) out of range")
    else:
        pyautogui.vscroll(clicks=clicks)


def pause(seconds):
    time.sleep(seconds)


# --------------------------------------------------------------------------
# ADVANCED FUNCTIONS
# --------------------------------------------------------------------------

def keep_awake(minutes=None):
    screen = pyautogui.size()
    safe_threshold = 25
    delay = 2
    if minutes is None:
        stop_condition = lambda: True
    elif isinstance(minutes, int) and 1 <= minutes < 1440:  # Between 1 minute and 1 day
        _end = time.time() + minutes * 60
        stop_condition = lambda: time.time() <= _end
    else:
        raise ValueError("Value provided for the parameter 'minutes' is out of range [1, 1440]")
    while stop_condition():
        # TODO: use my own click function
        pyautogui.moveTo(
            x=random.randint(safe_threshold, screen.width - safe_threshold),
            y=random.randint(safe_threshold, screen.height - safe_threshold))
        if random.randint(0, 3) == 1:  # 25% chances
            pyautogui.press('f5')
            time.sleep(delay)
            pyautogui.press('esc')
            time.sleep(delay)
        else:
            time.sleep(delay)


def quick_action(action):
    # TODO: Implement quick actions
    quick_actions = {
        'next_window': lambda: pyautogui.hotkey(*['alt', 'tab']),
        'next_tab': lambda: pyautogui.hotkey(*['ctrl', 'tab']),
        'close_tab': lambda: pyautogui.hotkey(*['ctrl', 'w']),
        'win_explorer': lambda: pyautogui.hotkey(*['win', 'e']),
        'cut': {'hold': ['ctrl'], 'do': ['x']},
        'open-link-new-tab': {'hold': ['ctrl'], 'do': [{'button': 'left', 'presses': 1, 'x': -1, 'y': -1}]},
        'copy': ['ctrl', 'c'],
        'paste': ['ctrl', 'v'],
    }
    # These are predefined quick actions
    # (right) click
    # double (right) click
    # triple (right) click
    # left click
    # middle click
    __PREDEFINED_ACTIONS = [
        {
            'command': 'press',
            'target': (None, None),
            'value': ['ctrl', 'c'],
            'description': 'copy'
        }
    ]
    pass


def with_hold_do(holds=None, actions=None):
    """
    We can define here quick actions. For example 'ctrl-click', 'ctrl-cmd', 'shift-cmd'

    Arguments
        holds: e.g. 'ctrl', 'shift', 'alt', 'alt-gr', 'fn'...
        actions: a list of actions
            keys 'a', 'b', 'c'...
            clicks {'clicks': 1, button: 'left', 'x': x, 'y': y}
                If (x, y) == (-1, -1) click in the current position
    """
    if holds is None:
        holds = []
    if actions is None:
        actions = []
    with pyautogui.hold(holds):
        for act in actions:
            if type(act) is tuple and len(act) == 2:
                click(x=act[0], y=act[1])
            elif type(act) is dict:
                click(button=act['button'], clicks=act['clicks'], x=act['x'], y=act['y'])
            elif type(act) is str:
                write(act)
            else:
                raise ValueError(f'Invalid type {type(act)} for hold_and_do(). ')


def copy_text(x=None, y=None, remove_special_chars=False):
    # TODO: refactor
    if x is None or y is None:
        x = pyautogui.position()[0]
        y = pyautogui.position()[1]
    # move(x, y)
    pyautogui.click(x=x, y=y, button='left', clicks=3)
    with_hold_do(['ctrl'], ['c'])
    if remove_special_chars:
        return re.sub('[\t\n\r]+', '', get_clipboard()).strip()
    else:
        return get_clipboard().strip()


def get_clipboard():
    # https://stackoverflow.com/questions/101128/how-do-i-read-text-from-the-clipboard
    # return Tk().clipboard_get()
    return pyperclip.paste()


def set_clipboard(data):
    # https://stackoverflow.com/questions/579687/how-do-i-copy-a-string-to-the-clipboard
    # TODO: fix. It does not work with Tkinter
    pyperclip.copy(str(data))


def get_coordinates():
    # TODO: Update according to the new class ScreenArea
    while True:
        elem_name = pyautogui.prompt('Place the cursor on the component and enter its name')
        if elem_name == '' or elem_name is None:
            break
        else:
            time.sleep(5)
            pos = pyautogui.position()
            print(f'{elem_name} = ({pos[0]}, {pos[1]})')


def find_window(target):
    for wd in pyautogui.getAllTitles():
        if re.search(target, wd) is not None:
            windows = pyautogui.getWindowsWithTitle(wd)
            if windows is None:
                raise FileNotFoundError(f"Window {target} not found. Is it open?")
            elif len(windows) != 1:
                raise ValueError(f"Ambiguous search. More than one window with name {target}")
            else:
                windows[0].activate()
                windows[0].maximize()
                return windows[0]
    raise FileNotFoundError(f"Window {target} not found. Is it open?")


# --------------------------------------------------------------------------
# CLASSES
# --------------------------------------------------------------------------

class Action:
    def __init__(self):
        self.command = ''  # click, move, write, press...
        self.target = ''  # (x,y)
        self.value = ''  # Additional args, e.g. [right, triple] -> click
        self.description = ''  # Optional short description of the action.


class Worker:
    __MAX_SPEED = 2.0
    __MIN_SPEED = 0.0

    def __init__(self, speed=None):
        _allowed = ["autobot", "millennial", "grandpa"]
        if re.match('[0-9]+(.[0-9])?', speed) is not None:
            speed = float(speed)
            if self.__MAX_SPEED < speed or speed < self.__MIN_SPEED:
                raise ValueError('Invalid speed range for the Worker')
        elif type(speed) is str:
            if speed in _allowed:
                speed = float(_allowed.index(speed))
            else:
                raise ValueError(f'Unknown Worker speed profile "{speed}". '
                                 f'Supported profiles (from slow to fast) are: {", ".join(_allowed)}')

        self.speed_profile = speed
        self._interval_delay = 0.0
        self.busy = False  # Indicate if the worker is currently executing a subroutine
        self.target_window = ''
        self.gui_layout = guiLayout.GUILayout()
        # STATS
        self.traveling_time = 0
        self.total_clicks = 0
        self.total_keys_pressed = 0

        self.update_speed_profile(0)
        keyboard.add_hotkey('ctrl+alt+add', lambda: self.update_speed_profile(0.5))
        keyboard.add_hotkey('ctrl+alt+subtract', lambda: self.update_speed_profile(-0.5))
        keyboard.add_hotkey('ctrl+alt+p', self.halt_resume_execution)

    def update_speed_profile(self, inc):
        if 0.0 <= self.speed_profile + float(inc) <= 2.0:
            self.speed_profile += float(inc)
            # TODO: this formula may be improved
            self._interval_delay = (0.05 + self.speed_profile / 10) * (self.speed_profile + 1)
            pyautogui.PAUSE = self.speed_profile
            print(f"[INFO] Speed profile updated (fastest <= current <= slowest): 0 <= {self.speed_profile} <= 2")
        elif self.speed_profile == 0.0:
            print('[WARN] Already using the slower speed profile')
        elif self.speed_profile == 2.0:
            print('[WARN] Already using the fastest speed profile')
        else:
            raise ValueError("Error. Speed profile out of bounds")

    def _pre_actions_checks(self):
        self.find_target_window()
        if not self.busy:
            exit(0)

    def is_target_window(self, alert=False):
        if re.search(self.target_window, pyautogui.getActiveWindow().title) is not None:
            return True
        else:
            if alert:
                res = pyautogui.confirm('WINDOW OUT OF SCOPE. Press enter to continue...')
                if res == 'Cancel':
                    print('[INFO] Operation cancelled by the user. Exiting...')
                    exit(0)
            return False

    def find_target_window(self):
        if self.is_target_window():
            pyautogui.getActiveWindow().maximize()
        else:
            find_window(self.target_window)

    def halt_resume_execution(self):
        self.busy = not self.busy
        # ret = pyautogui.confirm(title='Halt execution', text='Execution halted. Do you want to continue?')
        # if ret == 'OK':
        #     p.resume()
        #     return
        # elif ret == 'Cancel':
        #     p.terminate()
        #     sys.exit(0)
        # else:
        #     raise ValueError('Unexpected error in function halt()')
        signal.raise_signal(signal.SIGTERM)
        raise KeyboardInterrupt

    def execute(self, **kwargs):
        # TODO: wrapper for every action execution, so that if busy is set to false, it will stop immediately
        self._pre_actions_checks()
        while self.busy:
            pass

    def click(self, x=None, y=None, button='left', clicks=1):
        self.busy = True
        self._pre_actions_checks()
        click(x=x, y=y, button=button, clicks=clicks, interval=self._interval_delay)
        self.busy = False

    def move(self, x, y):
        self.busy = True
        self._pre_actions_checks()
        self.traveling_time += self.speed_profile
        move(x, y, self.speed_profile)
        self.busy = False

    def move_to(self, point: guiLayout.Point):
        self.busy = True
        self._pre_actions_checks()
        self.traveling_time += self.speed_profile
        pyautogui.moveTo(point.x, point.y, self.speed_profile, pyautogui.easeOutQuad)
        self.busy = False

    def write(self, msg):
        """
        Tries to write wherever the cursor is placed

        Arguments
            msg. String to be written. Can contain special characters (\n, \t...)
            If msg is a list of strings, then each element of the list will be written in a new line
        """
        self.busy = True
        self._pre_actions_checks()
        if type(msg) is str:
            write(msg=msg, interval=self._interval_delay / 2)
            self.total_keys_pressed += len(msg)
        elif type(msg) is list:
            for m in msg:
                write(msg=f'{m}\n', interval=self._interval_delay / 2)
                self.total_keys_pressed += len(m) + 1  # len(m) + \n
        else:
            raise TypeError('Invalid type for msg in write()')
        self.busy = False

    def press(self, *args):
        """
        press('ctrl', 'c') = press(*['ctrl', 'c']) = press('ctrl+c')
        You can pass a list of key-actions that will be executed in order, for example: 'ctrl+c', 'ctrl+v'
        Arguments (2 types)
            string
                 'a' -> single press
                 'a+b+c' -> hotkey (use \\ to escape +, like 3\\+2 = 5)
            tuple
                ('c', 2) -> {'key': 'c', 'presses': 2}
                ('ctrl+t', 2) -> {'key': 'ctrl+t', 'presses': 2}
        """
        self.busy = True
        self._pre_actions_checks()
        for arg in args:
            if type(arg) is tuple:
                _key = arg[0]
                _presses = arg[1]
            elif type(arg) is str:
                _key = arg
                _presses = 1
            else:
                raise TypeError('Invalid args type in press() ')

            if _key.find('+') != -1 and re.search('^\\w+(\\+\\w+)+', arg) is not None:
                for _i in range(0, _presses):
                    pyautogui.hotkey(*_key.split('+'))
                self.total_keys_pressed += len(_key.split('+')) * _presses
            else:
                pyautogui.press(keys=_key, presses=_presses, interval=self._interval_delay)
                self.total_keys_pressed += _presses
        self.busy = False

    def pause(self, seconds):
        self.traveling_time += seconds
        pause(seconds)

    # --------------------------------------------------------------------------
    # ADVANCED FUNCTIONS
    # --------------------------------------------------------------------------

    def keep_awake(self, minutes=None):
        self.busy = True
        if minutes is None:
            stop_condition = True
        elif isinstance(minutes, int) and 1 <= minutes < 1440:  # Between 1 minute and 1 day
            _end = time.time() + minutes * 60
            stop_condition = lambda: time.time() <= _end
        else:
            raise ValueError("Value provided for the parameter 'minutes' is out of range [1, 1440]")
        while stop_condition:
            move(x=self.gui_layout.screen_center()[0] + random.randint(-200, 200),
                 y=self.gui_layout.screen_center()[1] + random.randint(-200, 200))
            press('f5', 'esc')
            time.sleep(2)

    def quick_action(self, action):
        # TODO: Implement quick actions
        pass

    def with_hold_do(self, holds=None, actions=None):
        """
        We can define here quick actions. For example 'ctrl-click', 'ctrl-cmd', 'shift-cmd'

        Arguments
            holds: e.g. 'ctrl', 'shift', 'alt', 'alt-gr', 'fn'...
            actions: a list of actions
                keys 'a', 'b', 'c'...
                clicks {'clicks': 1, button: 'left', 'x': x, 'y': y}
                    If (x, y) == (-1, -1) click in the current position
        """
        self.busy = True
        self._pre_actions_checks()
        if holds is None:
            holds = []
        if actions is None:
            actions = []
        with pyautogui.hold(holds):
            for act in actions:
                if type(act) is tuple and len(act) == 2:
                    self.click(x=act[0], y=act[1])
                    self.total_clicks += 1
                elif type(act) is dict:
                    self.click(button=act['button'], clicks=act['clicks'], x=act['x'], y=act['y'])
                elif type(act) is str:
                    self.write(act)
                    self.total_keys_pressed += 1
                else:
                    raise ValueError(f'Invalid type {type(act)} for hold_and_do(). ')
                self.pause(self.speed_profile)
        self.busy = False


# -----------------------------------------------------------------------------------------------------------------
# ------------------------------------- MAIN-----------------------------------------------------------------------
# -----------------------------------------------------------------------------------------------------------------

if __name__ == '__main__':
    keep_awake()
    # w = Worker('millennial')
    # print(w.__dict__)
    # w.keep_awake(minutes=1)
    # exit(0)



import requests


class APIworker:
    def __init__(self, methods):
        __HTTP_METHODS__ = ['GET', 'HEAD', 'POST', 'PUT', 'DELETE', 'CONNECT', 'OPTIONS', 'TRACE', 'PATCH']
        if methods is None:
            methods = ['GET']
        self.consecutive_failures = 0
        self.statistics = {
            'executionTime': 0,
            'numOfRequests': 0,
            'consumedBandwidth': 0
        }
        self._allowed_methods = methods
        self.force_https = False
        self.base_url = 'https://blabla.com'
        self.endpoints = {
            'cart': 'cart',
            'products': 'products/{productId}?param1=value1&param2=value2'
        }
        # TODO: Wrapper for params
        # Opt 1. Vanilla string -> 'products': 'products/{productId}?param1=value1&param2=value2'
        # requests.get(f"{self.sec_problems_url}?{'&'.join(_query_params)}", headers=_headers).json()
        # Opt 2. Structured data
        endpoints = [
            {
                'endpoint': 'entities',
                'method': 'GET',
                'path_params': {
                    'entityID'
                },
                'request_params': {
                    'param1': 'value1',
                    'param2': 'value2'
                }
            },
            {
                'endpoint': 'securityProblems',
                'path_params': {
                    'entityID'
                },
                'request_params': {
                    'param1': 'value1',
                    'param2': 'value2'
                }
            }
        ]

    def request(self, url, method='GET', params=None, headers=None, cookies=None, verify=True):
        # TODO: develop encode params and url
        # TODO: review authentication, files... and the rest of parameters
        req_args = {'method': method, 'url': url, 'params': params,
                    'headers': headers, 'cookies': cookies, 'verify': verify}
        if method in self._allowed_methods:
            response = requests.request(**req_args)
        else:
            raise requests.HTTPError(f"Invalid request method {method}. Allowed: {', '.join(self._allowed_methods)}")

        if response.status_code == 200:
            return response
        else:
            raise requests.HTTPError(f'Unexpected status code ({response.status_code}) for the request {req_args}')

    def scrap(self, url, depth=0):
        from bs4 import BeautifulSoup

        w = APIworker.APIworker('GET')

        all_links = []
        MAX_DEPTH = 3
        if depth >= MAX_DEPTH:
            raise AttributeError("[ERROR] Max depth exceeded")
        res = w.request(url)
        soup = BeautifulSoup(res.text, 'html.parser')
        # print(soup.prettify())
        # do things. For example
        for link in soup.find_all('a'):
            href = link.get('href')[2:]
            all_links.append(href)



import klsframe.utilities.utils as utils


def test_dict_to_table():
    a = {'Nombre': 'pedro', 'Apellidos': 'perez', 'Edad': 12, 'hobbies': ['skateboarding']}
    print('\n-----> Compact mode <-----\n')
    print(utils.dict_to_table(a, compact=True))
    print('\n-----> Extended mode <-----\n')
    print(utils.dict_to_table(a))


def test_join_by():
    listilla = ['Pablito', 'clavo', 3, 'clavitos']
    print(utils.join_by(' + ', listilla))
    print(utils.join_by(' + ', [1, 2, 3, 4, 5, 6]))
    exit(0)


def test_sort_dict():
    dicto = {'1': 'shit', '2': 'here', '3': 'we', '4': 'go', '5': 'again'}
    print(f"Alphabetical Key order:\n{utils.sort_dict(dicto)}")
    print(f"Alphabetical inverse Key order:\n{utils.sort_dict(dicto, reverse=True)}")
    print(f"Alphabetical value order:\n{utils.sort_dict(dicto, sortby='values')}")
    print(f"Alphabetical inverse value order:\n{utils.sort_dict(dicto, sortby='values', reverse=True)}")


if __name__ == '__main__':
    test_dict_to_table()



# '''This module implements specialized container datatypes providing
# alternatives to Python's general purpose built-in containers, dict,
# list, set, and tuple.
#
# * namedtuple   factory function for creating tuple subclasses with named fields
# * deque        list-like container with fast appends and pops on either end
# * ChainMap     dict-like class for creating a single view of multiple mappings
# * Counter      dict subclass for counting hashable objects
# * OrderedDict  dict subclass that remembers the order entries were added
# * defaultdict  dict subclass that calls a factory function to supply missing values
# * UserDict     wrapper around dictionary objects for easier dict subclassing
# * UserList     wrapper around list objects for easier list subclassing
# * UserString   wrapper around string objects for easier string subclassing
#
# '''
#
# __version__ = '0.0.3'
# __all__ = [
#     'workers'
#     'ChainMap',
#     'Counter',
#     'OrderedDict',
#     'UserDict',
#     'UserList',
#     'UserString',
#     'defaultdict',
#     'deque',
#     'namedtuple',
# ]
# __author__ = 'Ramón Hernández León <rhdezl05@gmail.com>'
#
# import workers as workers
# import cli as cli
# import gui as gui
# import utilities as utilities
#
# # try:
# #     from _collections import deque
# # except ImportError:
# #     pass
# # else:
# #     _collections_abc.MutableSequence.register(deque)
#
# # def ChainMap():
# #     pass


from typing import Union

import pandas as pd
import re


def get_df_rows(df, rindex=None):
    """
    Get the specified rows of a Pandas data frame
    :param df: Pandas dataframe to be scrapped
    :param rindex: Index of the rows to retrieve. It accepts strings and integer numbers. Accepted:
        - None (default): all the rows are retrieved
        - 1: single number, single index
        - 1-4: indexes from 1 to 4, included
        - 2,3,4: indexes 2,3,4
        - 2,3,7-9: combination of the two previous modes
        - head-N: first N rows
        - tail-N: last N rows
        - first: acronym for head-1
        - last: acronym for tail-1
    :return: a list containing one dict per each row (the dict's keys are the column names, its values the row values)
    """
    data = []
    if rindex is None:
        data = []
        for index, row in df.iterrows():
            data.append({k: v for k, v in row.items()})
    elif isinstance(rindex, int):
        data = []
        for index, row in df.iterrows():
            if index == rindex:
                # data.append({k: v for k, v in row.items()})
                return {k: v for k, v in row.items()}
    else:
        rindex = str(rindex)
        if rindex == 'first':
            return get_df_rows(df, df.first_valid_index())
        elif rindex == 'last':
            return get_df_rows(df, df.last_valid_index())
        elif re.fullmatch('head-[0-9]+', rindex):
            rindex = range(df.first_valid_index(), int(rindex.split('-')[1]))
        elif re.fullmatch('tail-[0-9]+', rindex):
            rindex = range(df.last_valid_index() + 1 - int(rindex.split('-')[1]), df.last_valid_index() + 1)
        elif re.fullmatch('[0-9]+-[0-9]+', rindex):
            rindex = range(int(rindex.split('-')[0]), int(rindex.split('-')[1]) + 1)
        elif re.fullmatch('([0-9](-[0-9]+)?,)*[0-9]+(-[0-9]+)?', rindex):
            # rindex = [int(i) for i in rindex.split(',')]
            aux = []
            for i in rindex.split(','):
                try:
                    aux.append(int(i))
                except ValueError:
                    aux.extend(range(int(i.split('-')[0]), int(i.split('-')[1]) + 1))
            rindex = aux
        else:
            raise ValueError("Unexpected value for parameter 'rindex'")

        for index, row in df.iterrows():
            if index in rindex:
                data.append({k: v for k, v in row.items()})
    return data


def df_add_column(dataframe, col, on, how='left') -> pd.DataFrame:
    return dataframe.merge(col, on=on, how=how)


def df_to_excel(output_file, contents: Union[dict, pd.DataFrame]) -> None:
    """
    Saves the contents of a dataframe into an Excel file.

    contents_example = {
        'MasterNmap': df,
        'Metrics_Ports': df_statistics_master,
        'Metrics_Ports_2': df_IPs_StatusPorts,
        'Sum_OpenPorts': df_summary_openports,
        'Sum_ClosedPorts': df_summary_closedports,
        'SumFilteredPorts': df_summary_filteredports
    }
    :param output_file: name for the Excel file
    :param contents: single DataFrame or a dictionary,
        whose keys are page names, and its values the corresponding dataframe
    :return: None
    """
    if isinstance(contents, dict):
        with pd.ExcelWriter(f"{output_file}.xlsx") as writer:
            for sheet_name, dataframe in contents.items():
                dataframe.to_excel(writer, sheet_name=str(sheet_name), index=False, header=True)
    elif isinstance(contents, pd.DataFrame):
        with pd.ExcelWriter(f"{output_file}.xlsx") as writer:
            contents.to_excel(writer, index=False, header=True)
    else:
        raise TypeError("Invalid type for paramterer 'mapping'. Expected: dict|DataFrame")



from datetime import datetime

stats_example_json = {
    'total clicks': 0,
    'total_key_presses': 0,
    'time_traveling': 0
}


def timing(func, *args):
    start_dt = datetime.now()
    func(*args)
    elapsed = datetime.now() - start_dt
    print(
        f'[INFO] Execution completed {int(elapsed.seconds / 60)} '
        f'min {elapsed.seconds % 60} s {elapsed.microseconds} µs')


if __name__ == '__main__':
    # Do some quick tests here
    pass



import klsframe.gui.guiLayout as guiLayout
import os


def test_gui_layout_export_import():
    _file = 'tmp.json'
    dummy = [
        {'view_assets': {
            "name_col": (-1, -1),
            "secondary_id_col": (-1, -1),
            "ip_col": (-1, -1),
            "numof_comp_col": (-1, -1),
            "numof_vulns_col": (-1, -1),
            "edit_button": (-1, -1),
            "next_page_button": (-1, -1),
            "search_bar": (-1, -1),
            "delete_button": (-1, -1),
            "confirm_deletion_button": (-1, -1)
        }},
        {'asset_components': {
            "name_col": (-1, -1),
            "cpe_col": (-1, -1),
            "port_col": (-1, -1)
        }},
        {'asset_details': {
            "name_field": (-1, -1),
            "client_id_field": (-1, -1),
            "secondary_id_field": (-1, -1),
            "asset_type_field": (-1, -1),
            "ip_field": (-1, -1),
            "purpose_field": (-1, -1),
            "description_field": (-1, -1)
        }}
    ]
    gui = guiLayout.GUILayout().load(dummy)
    print("DEBUG 1", gui.to_json())
    gui.local_save(_file)
    gui2 = guiLayout.GUILayout.load(_file)
    print("DEBUG 2", gui2.to_json())
    os.remove(_file)
    assert gui == gui2


def test_guiLayout_table():
    # TODO: complete test_guiLayout_table()
    table = guiLayout.GUITable("Test table")
    table.table_size = 11
    table.row_height = 28
    table.first_row_Y = 235


if __name__ == '__main__':
    pass



import klsframe.utilities.statistics as stats
import time


def test_timing():
    def addition(*args):
        try:
            total = 0
            for arg in args:
                total += int(arg)
            print(f"{' + '.join([str(val) for val in args])} = {total}")
            time.sleep(3)
        except TypeError:
            print('[ERROR] Invalid parameters. Only numbers allowed')

    stats.timing(addition, 2, 3)


if __name__ == '__main__':
    test_timing()
